(function ($) {
    "use strict";
    function Grid(dimensions) {
        var i;
        this.dimensions = dimensions;
        this.cells = [].constructor(this.dimensions[0] * this.dimensions[1]);
        for (i = 0; i < this.cells.length; i += 1) {
            this.cells[i] = new Cell([ parseInt(i / dimensions[1], 10), i % dimensions[1]]);
        }
        //console.log(" created grid with dimension " + dimensions + " and " + this.cells.length + " cells");
        this.getDimensions = function () {
            return this.dimensions;
        };
        this.getCell = function (position) {
            if (position[1] > this.dimensions[1] - 1) {
                return null;
            }
            return this.cells[this.getID(position)];
        };
        this.getID = function (position) {
            return position[0] * this.dimensions[1] + position[1];
        };
        this.getItemList = function () {
            var i, cell, items = [];
            for (i = 0; i < this.getCells().length; i = i + 1) {
                cell = this.getCells()[i];
                if (cell.getItem() !== null) {
                    items[items.length] = cell.getItem();
                }
            }
            return items;
        };
        this.getSortedItemsIds = function () {
            var i, cell, itemids = [];
            for (i = 0; i < this.getCells().length; i += 1) {
                cell = this.getCells()[i];
                /*
                 if (cell.getItem() !== null) {
                 console.log(cell.getPosition() + ": " + cell.getItem().getId() );
                 }
                 */
                if (cell.getItem() !== null && itemids.indexOf(String(cell.getItem().getId())) < 0) {
                    itemids[itemids.length] = String(cell.getItem().getId());
                }
            }
            return itemids;
        };
        this.putItem = function (currentCell, item) {
            var cell, j, i, itemWidth = parseInt(item.getDimensions()[0], 10), itemHeight = parseInt(item.getDimensions()[1], 10);
            if (currentCell === null || currentCell === undefined) {
                //console.log("currentCell is not defined! ");
                return false;
            }
            //console.log("trying to place item in cell " + currentCell.getPosition());
            if (currentCell.getItem() !== null || item === null) {
                //console.log("cell is occupied or item is empty!");
                return false;
            }
            //check if item fits into space starting with this cell
            for (i = currentCell.getPosition()[1]; i < parseInt(currentCell.getPosition()[1], 10) + itemWidth; i += 1) {
                for (j = currentCell.getPosition()[0]; j < parseInt(currentCell.getPosition()[0], 10) + itemHeight; j += 1) {
                    cell = this.getCell([j, i]);
                    if (cell === null || cell === undefined || cell.getItem() !== null) {
                        return false;
                    }
                }
            }

            // found space for item => place item
            for (i = currentCell.getPosition()[1]; i < parseInt(currentCell.getPosition()[1], 10) + itemWidth; i += 1) {
                for (j = currentCell.getPosition()[0]; j < parseInt(currentCell.getPosition()[0], 10) + itemHeight; j += 1) {
                    cell = this.getCell([j, i]);
                    //console.log("found place for item " + item.getId() + ": " + j + "," + i + " size: " + itemWidth + ", " + itemHeight);
                    //console.log(j + ", " +  cell.getPosition() + ", " +  itemHeight + ", " + (j < cell.getPosition()[0] + itemHeight));
                    cell.setItem(item);
                }

            }
            return true;
        };
        this.getCells = function () {
            return this.cells;
        };
        this.getItemCells = function (id) {
            var itemCells = [], i, cell;
            for (i = 0; i < this.cells.length; i += 1) {
                cell = this.cells[i];
                if (cell.getItem() !== null && cell.getItem().getId() === id) {
                    itemCells[itemCells.length] = cell;
                }
            }
            return itemCells;
        };
        this.getFirstItemCell = function (pid) {
            var i, item;
            for (i = 0; i < this.cells.length; i += 1) {
                item = this.cells[i].getItem();
                if (item !== null && item.getId() === pid) {
                    return this.cells[i];
                }
            }
            return null;
        };
        this.reset = function () {
            var i;
            for (i = 0; i < this.getCells().length; i += 1) {
                this.getCells()[i].setItem(null);
            }
        };
    }
    function Cell(position) {
        this.setItem = function (item) {
            this.item = item;
        };
        this.getItem = function () {
            return this.item;
        };
        this.getPosition = function () {
            return this.position;
        };
        this.position = position;
        this.item = null;
        return true;
    }
    function Item(id, dimensions) {
        this.setDimensions = function (dimensions) {
            this.dimensions = dimensions;
        };
        this.getDimensions = function () {
            return this.dimensions;
        };
        this.getId = function () {
            return this.id;
        };
        this.dimensions = dimensions;
        this.id = id;
        return true;
    }
    var methods = {
        sortGrid: function (grid, items, sortsequence) {
            //clean grid
            var cells = grid.getCells(), i, j, cell, item;
            grid.reset();
            for (i = 0; i < sortsequence.length; i += 1) {
                // console.log(sortsequence[i]);
                item = items[sortsequence[i]];
                for (j = 0; j < cells.length; j += 1) {
                    //console.log("trying to place item " + item + " in cell " + j);
                    cell = cells[j];

                    /*                  if (cell.getItem() === null || cell.getItem() === undefined) {
                     console.log("cell is empty.");
                     } else {
                     console.log("cell is occupied.");
                     }
                     */
                    if (cell.getItem() === null && grid.putItem(cell, item)) {
                        //console.log("put item " + item.getId() + " in " + cell.getPosition() + " size: " + item.getDimensions());
                        break;
                    } /*else {
                     console.log(" not fit in " + cell.getPosition());
                     }*/
                }
            }
        },
        init: function (options) {

            // if no cellSize is given set cellSize to first element's outer width
            if (options.cellSize === null || options.cellSize === undefined) {
                options.cellSize = $(this).children(":first").outerWidth(true);
                //console.log("setting cell size to " + options.cellSize);
            }
            // if no inner cellSize is given set inner cellSize to first element's width
            if (options.innerCellSize === null || options.innerCellSize === undefined) {
                options.innerCellSize = $(this).children(":first").width();
                //console.log("setting innerCellSize to " + options.innerCellSize);
            }
            // merge with default options. This options are stored for each gridpacking separately
            options = $.extend({
                gridWidth     : 5,
                padding       : 0,
                cellSize      : 100,
                innerCellSize : 98,
                maxItemSize   : 4,
                defaultCellDimensions : [1, 1]
            }, options);

            // for each grid! (not for each item in the grid)
            return this.each(function () {
                var items, grid, $this = $(this), data = $this.data('gridpacking');
                // initialize items and grid
                items = methods.initializeItems($(this), options);
                grid = new Grid([items.length * options.maxItemSize, options.gridWidth]);

                // If the plugin hasn't been initialized yet save items and grid in data object
                if (!data) {
                    $(this).data('gridpacking', {
                        target: $this,
                        items: items,
                        grid: grid,
                        options: options
                    });
                }
                methods.rearrange($(this));
            });
        },
        rearrange: function (gridpacking) {
            // core-method: rearrange the products on grid and update the view
            var sequence, start, end, data, items, grid, options, totalHeight;
            data = gridpacking.data('gridpacking');
            items = data.items;
            grid = data.grid;
            options = data.options;
            sequence = methods.getSequence(gridpacking);
            // calculate new sequence
            // console.log("rearranging " + sequence);
            start = new Date().getTime();
            methods.sortGrid(grid, items, sequence);
            totalHeight = 0;
            // update the sortables (position,...)
            gridpacking.children().each(
                function () {
                    var item, pid, box, dim, pos;
                    item = $(this);
                    pid = Number(item.attr('data-gp-id'));
                    // console.log(" position item " + pid);
                    //get first (top,left) cell for item
                    box = grid.getFirstItemCell(pid);
                    //set class according to position
                    dim = items[pid].getDimensions();
                    if (typeof (box) !== 'undefined' && box !== null) {
                        //console.log("box: " + box.getPosition() + ", " + dim);
                        pos = box.getPosition();
                        if (pos[0] + dim[1] > totalHeight) {
                            totalHeight = pos[0] + dim[1];
                            //console.log("adjusting total height: " + totalHeight);
                        }

                        item.css({
                            left: pos[1] * (options.cellSize + options.padding),
                            top: pos[0] * (options.cellSize + options.padding),
                            width: dim[0] * options.cellSize - 2 * (options.cellSize - options.innerCellSize),
                            height: dim[1] * options.cellSize - 2 * (options.cellSize - options.innerCellSize),
                            visibility: 'visible'
                        });
                    }
                }
            );
            gridpacking.css("min-height", (totalHeight * options.cellSize) + 'px');
            console.log('Rearranging took: ' + (new Date().getTime() - start) + ' ms');
        },
        getSequence: function (sortable) {
            // helper method which returns the current sequence of the sortable as an array of integers,
            // i.e. an array containing the id's of the sortables (remove the string part at the beginning of the id)
            var idSequence = [];
            sortable.children().each(
                function () {
                    idSequence[idSequence.length] = Number($(this).attr('data-gp-id'));
                }
            );
            return idSequence;
        },
        initializeItems: function (itemGrid, options) {
            var items = [];
            itemGrid.children().each(function () {
                var item = $(this), width =  options.defaultCellDimensions[0], height = options.defaultCellDimensions[1];
                if (!isNaN(item.attr('data-gp-width')) && Number(item.attr('data-gp-width')) > 0) {
                    width = Number(item.attr('data-gp-width'));
                }
                if (!isNaN(item.attr('data-gp-height')) && Number(item.attr('data-gp-height')) > 0) {
                    height = Number(item.attr('data-gp-height'));
                }
                items[items.length] = new Item(Number(item.attr('data-gp-id')), [width, height]);
                //console.log('added item ' + item.attr('data-gp-id') + ' with dimension ' + items[items.length -1].dimensions);
            });
            return items;
        }
    };
    $.fn.gridpacking = function (method) {

        if (methods[method]) {
            return methods[method].apply(this, Array.prototype.slice.call(arguments, 1));
        } else if (typeof method === 'object' || !method) {
            return methods.init.apply(this, arguments);
        } else {
            $.error('Method ' + method + ' does not exist on jQuery.gridpacking');
        }

    };
})(jQuery);
